package java.time.chrono;

import java.time.temporal.Temporal;
import java.io.DataInput;
import java.io.IOException;
import java.io.DataOutput;
import java.io.InvalidObjectException;
import java.io.ObjectInputStream;
import java.time.Period;
import java.time.LocalTime;
import java.time.temporal.TemporalUnit;
import java.time.temporal.TemporalAmount;
import java.time.temporal.TemporalAdjuster;
import sun.util.calendar.Era;
import java.time.temporal.UnsupportedTemporalTypeException;
import java.time.temporal.ValueRange;
import java.time.temporal.ChronoField;
import java.time.temporal.TemporalField;
import java.util.Calendar;
import java.time.temporal.TemporalAccessor;
import sun.util.calendar.CalendarDate;
import sun.util.calendar.LocalGregorianCalendar;
import java.time.DateTimeException;
import java.util.Objects;
import java.time.ZoneId;
import java.time.Clock;
import java.time.LocalDate;
import java.io.Serializable;

public final class JapaneseDate extends ChronoLocalDateImpl<JapaneseDate> implements ChronoLocalDate, Serializable
{
    private static final long serialVersionUID = -305327627230580483L;
    private final transient LocalDate isoDate;
    private transient JapaneseEra era;
    private transient int yearOfEra;
    static final LocalDate MEIJI_6_ISODATE;
    
    public static JapaneseDate now() {
        return now(Clock.systemDefaultZone());
    }
    
    public static JapaneseDate now(ZoneId zone) {
        return now(Clock.system(zone));
    }
    
    public static JapaneseDate now(Clock clock) {
        return new JapaneseDate(LocalDate.now(clock));
    }
    
    public static JapaneseDate of(JapaneseEra obj, int year, int n, int n2) {
        Objects.requireNonNull(obj, "era");
        LocalGregorianCalendar.Date calendarDate = JapaneseChronology.JCAL.newCalendarDate(null);
        calendarDate.setEra(obj.getPrivateEra()).setDate(year, n, n2);
        if (!JapaneseChronology.JCAL.validate(calendarDate)) {
            throw new DateTimeException("year, month, and day not valid for Era");
        }
        return new JapaneseDate(obj, year, LocalDate.of(calendarDate.getNormalizedYear(), n, n2));
    }
    
    public static JapaneseDate of(int year, int month, int dayOfMonth) {
        return new JapaneseDate(LocalDate.of(year, month, dayOfMonth));
    }
    
    static JapaneseDate ofYearDay(JapaneseEra obj, int n, int dayOfMonth) {
        Objects.requireNonNull(obj, "era");
        CalendarDate sinceDate = obj.getPrivateEra().getSinceDate();
        LocalGregorianCalendar.Date calendarDate = JapaneseChronology.JCAL.newCalendarDate(null);
        calendarDate.setEra(obj.getPrivateEra());
        if (n == 1) {
            calendarDate.setDate(n, sinceDate.getMonth(), sinceDate.getDayOfMonth() + dayOfMonth - 1);
        }
        else {
            calendarDate.setDate(n, 1, dayOfMonth);
        }
        JapaneseChronology.JCAL.normalize(calendarDate);
        if (obj.getPrivateEra() != calendarDate.getEra() || n != calendarDate.getYear()) {
            throw new DateTimeException("Invalid parameters");
        }
        return new JapaneseDate(obj, n, LocalDate.of(calendarDate.getNormalizedYear(), calendarDate.getMonth(), calendarDate.getDayOfMonth()));
    }
    
    public static JapaneseDate from(TemporalAccessor temporalAccessor) {
        return JapaneseChronology.INSTANCE.date(temporalAccessor);
    }
    
    JapaneseDate(LocalDate isoDate) {
        if (isoDate.isBefore(JapaneseDate.MEIJI_6_ISODATE)) {
            throw new DateTimeException("JapaneseDate before Meiji 6 is not supported");
        }
        LocalGregorianCalendar.Date privateJapaneseDate = toPrivateJapaneseDate(isoDate);
        this.era = JapaneseEra.toJapaneseEra(privateJapaneseDate.getEra());
        this.yearOfEra = privateJapaneseDate.getYear();
        this.isoDate = isoDate;
        return;
    }
    
    JapaneseDate(JapaneseEra era, int yearOfEra, LocalDate isoDate) {
        if (isoDate.isBefore(JapaneseDate.MEIJI_6_ISODATE)) {
            throw new DateTimeException("JapaneseDate before Meiji 6 is not supported");
        }
        this.era = era;
        this.yearOfEra = yearOfEra;
        this.isoDate = isoDate;
        return;
    }
    
    @Override
    public JapaneseChronology getChronology() {
        return JapaneseChronology.INSTANCE;
    }
    
    @Override
    public JapaneseEra getEra() {
        return this.era;
    }
    
    @Override
    public int lengthOfMonth() {
        return this.isoDate.lengthOfMonth();
    }
    
    @Override
    public int lengthOfYear() {
        Calendar instance = Calendar.getInstance(JapaneseChronology.LOCALE);
        instance.set(0, this.era.getValue() + 2);
        instance.set(this.yearOfEra, this.isoDate.getMonthValue() - 1, this.isoDate.getDayOfMonth());
        return instance.getActualMaximum(6);
    }
    
    @Override
    public boolean isSupported(TemporalField temporalField) {
        return temporalField != ChronoField.ALIGNED_DAY_OF_WEEK_IN_MONTH && temporalField != ChronoField.ALIGNED_DAY_OF_WEEK_IN_YEAR && temporalField != ChronoField.ALIGNED_WEEK_OF_MONTH && temporalField != ChronoField.ALIGNED_WEEK_OF_YEAR && super.isSupported(temporalField);
    }
    
    @Override
    public ValueRange range(TemporalField obj) {
        if ((obj instanceof ChronoField)) {
            if (this.isSupported(obj)) {
                ChronoField chronoField = (ChronoField)obj;
                switch (chronoField) {
                    case DAY_OF_MONTH: {
                        return ValueRange.of(1L, (long)this.lengthOfMonth());
                    }
                    case DAY_OF_YEAR: {
                        return ValueRange.of(1L, (long)this.lengthOfYear());
                    }
                    case YEAR_OF_ERA: {
                        Calendar instance = Calendar.getInstance(JapaneseChronology.LOCALE);
                        instance.set(0, this.era.getValue() + 2);
                        instance.set(this.yearOfEra, this.isoDate.getMonthValue() - 1, this.isoDate.getDayOfMonth());
                        return ValueRange.of(1L, (long)instance.getActualMaximum(1));
                    }
                    default: {
                        return this.getChronology().range(chronoField);
                    }
                }
            }
            else {
                throw new UnsupportedTemporalTypeException("Unsupported field: " + obj);
            }
        }
        else {
            return obj.rangeRefinedBy(this);
        }
    }
    
    @Override
    public long getLong(TemporalField temporalField) {
        if ((temporalField instanceof ChronoField)) {
            switch ((ChronoField)temporalField) {
                case ALIGNED_DAY_OF_WEEK_IN_MONTH:
                case ALIGNED_DAY_OF_WEEK_IN_YEAR:
                case ALIGNED_WEEK_OF_MONTH:
                case ALIGNED_WEEK_OF_YEAR: {
                    throw new UnsupportedTemporalTypeException("Unsupported field: " + temporalField);
                }
                case YEAR_OF_ERA: {
                    return (long)this.yearOfEra;
                }
                case ERA: {
                    return (long)this.era.getValue();
                }
                case DAY_OF_YEAR: {
                    Calendar instance = Calendar.getInstance(JapaneseChronology.LOCALE);
                    instance.set(0, this.era.getValue() + 2);
                    instance.set(this.yearOfEra, this.isoDate.getMonthValue() - 1, this.isoDate.getDayOfMonth());
                    return (long)instance.get(6);
                }
                default: {
                    return this.isoDate.getLong(temporalField);
                }
            }
        }
        else {
            return temporalField.getFrom(this);
        }
    }
    
    private static LocalGregorianCalendar.Date toPrivateJapaneseDate(LocalDate localDate) {
        LocalGregorianCalendar.Date calendarDate = JapaneseChronology.JCAL.newCalendarDate(null);
        Era privateEra = JapaneseEra.privateEraFrom(localDate);
        int year = localDate.getYear();
        if (privateEra != null) {
            year -= privateEra.getSinceDate().getYear() - 1;
        }
        calendarDate.setEra(privateEra).setYear(year).setMonth(localDate.getMonthValue()).setDayOfMonth(localDate.getDayOfMonth());
        JapaneseChronology.JCAL.normalize(calendarDate);
        return calendarDate;
    }
    
    @Override
    public JapaneseDate with(TemporalField field, long n) {
        if ((field instanceof ChronoField)) {
            ChronoField field2 = (ChronoField)field;
            if (this.getLong(field2) == n) {
                return this;
            }
            switch (field2) {
                case YEAR_OF_ERA:
                case ERA:
                case YEAR: {
                    int checkValidIntValue = this.getChronology().range(field2).checkValidIntValue(n, field2);
                    switch (field2) {
                        case YEAR_OF_ERA: {
                            return this.withYear(checkValidIntValue);
                        }
                        case YEAR: {
                            return this.with(this.isoDate.withYear(checkValidIntValue));
                        }
                        case ERA: {
                            return this.withYear(JapaneseEra.of(checkValidIntValue), this.yearOfEra);
                        }
                        default: {
                            break;
                        }
                    }
                    break;
                }
            }
            return this.with(this.isoDate.with(field, n));
        }
        else {
            return (JapaneseDate)super.with(field, n);
        }
    }
    
    @Override
    public JapaneseDate with(TemporalAdjuster temporalAdjuster) {
        return (JapaneseDate)super.with(temporalAdjuster);
    }
    
    @Override
    public JapaneseDate plus(TemporalAmount temporalAmount) {
        return (JapaneseDate)super.plus(temporalAmount);
    }
    
    @Override
    public JapaneseDate minus(TemporalAmount temporalAmount) {
        return (JapaneseDate)super.minus(temporalAmount);
    }
    
    private JapaneseDate withYear(JapaneseEra japaneseEra, int n) {
        return this.with(this.isoDate.withYear(JapaneseChronology.INSTANCE.prolepticYear(japaneseEra, n)));
    }
    
    private JapaneseDate withYear(int n) {
        return this.withYear(this.getEra(), n);
    }
    
    @Override
    JapaneseDate plusYears(long yearsToAdd) {
        return this.with(this.isoDate.plusYears(yearsToAdd));
    }
    
    @Override
    JapaneseDate plusMonths(long monthsToAdd) {
        return this.with(this.isoDate.plusMonths(monthsToAdd));
    }
    
    @Override
    JapaneseDate plusWeeks(long weeksToAdd) {
        return this.with(this.isoDate.plusWeeks(weeksToAdd));
    }
    
    @Override
    JapaneseDate plusDays(long daysToAdd) {
        return this.with(this.isoDate.plusDays(daysToAdd));
    }
    
    @Override
    public JapaneseDate plus(long n, TemporalUnit temporalUnit) {
        return (JapaneseDate)super.plus(n, temporalUnit);
    }
    
    @Override
    public JapaneseDate minus(long n, TemporalUnit temporalUnit) {
        return (JapaneseDate)super.minus(n, temporalUnit);
    }
    
    @Override
    JapaneseDate minusYears(long n) {
        return (JapaneseDate)super.minusYears(n);
    }
    
    @Override
    JapaneseDate minusMonths(long n) {
        return (JapaneseDate)super.minusMonths(n);
    }
    
    @Override
    JapaneseDate minusWeeks(long n) {
        return (JapaneseDate)super.minusWeeks(n);
    }
    
    @Override
    JapaneseDate minusDays(long n) {
        return (JapaneseDate)super.minusDays(n);
    }
    
    private JapaneseDate with(LocalDate localDate) {
        return localDate.equals(this.isoDate) ? this : new JapaneseDate(localDate);
    }
    
    @Override
    public final ChronoLocalDateTime<JapaneseDate> atTime(LocalTime localTime) {
        return super.atTime(localTime);
    }
    
    @Override
    public ChronoPeriod until(ChronoLocalDate endDateExclusive) {
        Period until = this.isoDate.until(endDateExclusive);
        return this.getChronology().period(until.getYears(), until.getMonths(), until.getDays());
    }
    
    @Override
    public long toEpochDay() {
        return this.isoDate.toEpochDay();
    }
    
    @Override
    public boolean equals(Object o) {
        return this == o || (o instanceof JapaneseDate && this.isoDate.equals(((JapaneseDate)o).isoDate));
    }
    
    @Override
    public int hashCode() {
        return this.getChronology().getId().hashCode() ^ this.isoDate.hashCode();
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws InvalidObjectException {
        throw new InvalidObjectException("Deserialization via serialization delegate");
    }
    
    private Object writeReplace() {
        return new Ser((byte)4, this);
    }
    
    void writeExternal(DataOutput dataOutput) throws IOException {
        dataOutput.writeInt(this.get(ChronoField.YEAR));
        dataOutput.writeByte(this.get(ChronoField.MONTH_OF_YEAR));
        dataOutput.writeByte(this.get(ChronoField.DAY_OF_MONTH));
    }
    
    static JapaneseDate readExternal(DataInput dataInput) throws IOException {
        return JapaneseChronology.INSTANCE.date(dataInput.readInt(), (int)dataInput.readByte(), (int)dataInput.readByte());
    }
    
    static {
        MEIJI_6_ISODATE = LocalDate.of(1873, 1, 1);
    }
}
